""" Generator functions related to trial and test functions and the accumulation loop"""

from dune.perftool.generation import domain, iname, pymbolic_expr, symbol, globalarg
from dune.perftool.ufl.modified_terminals import ModifiedArgumentDescriptor
from dune.perftool.pdelab import (name_index,
                                  restricted_name,
                                  )
from dune.perftool.pdelab.basis import (evaluate_trialfunction,
                                        evaluate_trialfunction_gradient,
                                        lfs_iname,
                                        name_basis,
                                        name_basis_gradient,
                                        name_lfs_bound,
                                        )
from dune.perftool import Restriction
from pymbolic.primitives import Subscript, Variable


@symbol
def name_testfunction_gradient(ma):
    assert ma.grad

    return name_basis_gradient(ma.argexpr.element(), ma.restriction)


@symbol
def name_testfunction(ma):
    assert not ma.grad

    return name_basis(ma.argexpr.element(), ma.restriction)


@pymbolic_expr
def pymbolic_testfunction(ma):
    # we only accept an index if we treat a gradient...
    assert bool(ma.index) == ma.grad

    if ma.grad:
        return Subscript(Variable(name_testfunction_gradient(ma)), (Variable(lfs_iname(ma.argexpr.element())), Variable(name_index(ma.index))))
    else:
        return Subscript(Variable(name_testfunction(ma)), Variable(lfs_iname(ma.argexpr.element())))


@symbol
def name_trialfunction_gradient(ma):
    assert ma.grad

    name = restricted_name("gradu", ma.restriction)
    evaluate_trialfunction_gradient(ma.argexpr.element(), name, ma.restriction)
    return name


@symbol
def name_trialfunction(ma):
    assert not ma.grad

    name = restricted_name("u", ma.restriction)
    evaluate_trialfunction(ma.argexpr.element(), name, ma.restriction)
    return name


@pymbolic_expr
def pymbolic_trialfunction(ma):
    # we only accept an index if we treat a gradient...
    assert bool(ma.index) == ma.grad

    if ma.grad:
        return Subscript(Variable(name_trialfunction_gradient(ma)), Variable(name_index(ma.index)))
    else:
        return Variable(name_trialfunction(ma))


@symbol
def name_testfunctionspace(restriction):
    return restricted_name("lfsv", restriction)


@symbol
def name_trialfunctionspace(restriction):
    return restricted_name("lfsu", restriction)


@symbol
def type_testfunctionspace():
    return "LFSV"


@symbol
def type_trialfunctionspace():
    return "LFSU"


@symbol
def name_coefficientcontainer(restriction):
    return restricted_name("x", restriction)


@symbol
def type_coefficientcontainer():
    return "X"


def name_argumentspace(ma):
    from ufl.classes import Argument, Coefficient
    if isinstance(ma.argexpr, Argument):
        if ma.argexpr.number() == 0:
            return name_testfunctionspace(ma.restriction)
        if ma.argexpr.number() == 1:
            return name_trialfunctionspace(ma.restriction)
    if isinstance(ma.argexpr, Coefficient):
        assert ma.argexpr.count() == 0
        return name_trialfunctionspace(ma.restriction)
    # We should never encounter an argument other than 0 or 1
    assert False


def name_argument(ma):
    if ma.argexpr.number() == 0:
        return name_testfunction(ma)
    if ma.argexpr.number() == 1:
        return name_trialfunction(ma)
    # We should never encounter an argument other than 0 or 1
    assert False


def pymbolic_argument(ma):
    if ma.argexpr.number() == 0:
        return pymbolic_testfunction(ma)
    if ma.argexpr.number() == 1:
        return pymbolic_trialfunction(ma)
    assert False


@symbol
def name_jacobian(restriction1, restriction2):
    # Restrictions may only differ if NONE
    if (restriction1 == Restriction.NONE) or (restriction2 == Restriction.NONE):
        assert restriction1 == restriction2

    return restricted_name(restricted_name("jac", restriction1), restriction2)


@symbol
def type_jacobian():
    return "J"


@symbol
def name_residual(restriction):
    return restricted_name("r", restriction)


@symbol
def type_residual():
    return "R"


def name_accumulation_variable(restrictions=(Restriction.NONE,)):
    from dune.perftool.generation import get_global_context_value
    ft = get_global_context_value("form_type")
    if ft == 'residual':
        return name_residual(*restrictions)
    if ft == 'jacobian':
        return name_jacobian(*restrictions)
    assert False


def type_accumulation_variable():
    from dune.perftool.generation import get_global_context_value
    ft = get_global_context_value("form_type")
    if ft == 'residual':
        return type_residual()
    if ft == 'jacobian':
        return type_jacobian()
    assert False
