""" Generators for parameter functions """

from dune.perftool.generation import (class_basename,
                                      class_member,
                                      constructor_parameter,
                                      generator_factory,
                                      initializer_list,
                                      preamble,
                                      symbol,
                                      temporary_variable
                                      )
from dune.perftool.pdelab.geometry import (name_cell,
                                           name_intersection,
                                           )
from dune.perftool.pdelab.quadrature import (name_quadrature_position,
                                             name_quadrature_position_in_cell,
                                             quadrature_preamble,
                                             )
from dune.perftool.cgen.clazz import AccessModifier
from dune.perftool.pdelab.localoperator import (class_type_from_cache,
                                                localoperator_basename,
                                                )


@class_basename("parameterclass")
def parameterclass_basename():
    lopbase = localoperator_basename()
    return "{}Params".format(lopbase)


@class_member("operator")
def define_parameterclass(name):
    _, t = class_type_from_cache("parameterclass")
    constructor_parameter("const {}&".format(t), name + "_", classtag="operator")
    initializer_list(name, [name + "_"], classtag="operator")
    return "const {}& {};".format(t, name)


@symbol
def name_paramclass():
    define_parameterclass("param")
    return "param"


@class_member("parameterclass", access=AccessModifier.PUBLIC)
def define_parameter_function_class_member(name, expr, t, cell):
    geot = "E" if cell else "I"
    geo = geot.lower()
    result = ["template<typename {}, typename X>".format(geot),
              "{} {}(const {}& {}, const X& local) const".format(t, name, geot, geo),
              "{",
              ]

    if expr.is_global:
        result.append("  auto x = {}.geometry().global(local);".format(geo))
    else:
        result.append("  auto x = local;")

    result.append("  " + expr.c_expr)
    result.append("}")

    return result


def evaluate_cell_parameter_function(name, restriction):
    param = name_paramclass()
    entity = name_cell(restriction)
    pos = name_quadrature_position_in_cell(restriction)
    return quadrature_preamble('{} = {}.{}({}, {});'.format(name,
                                                            name_paramclass(),
                                                            name,
                                                            entity,
                                                            pos,
                                                            ),
                               assignees=frozenset({name}),
                               read_variables=frozenset({pos}),
                               )


def evaluate_intersection_parameter_function(name):
    # Check that this is not a volume term, as that would not be well-defined
    from dune.perftool.generation import get_global_context_value
    it = get_global_context_value("integral_type")
    assert it is not 'cell'

    param = name_paramclass()
    intersection = name_intersection()
    pos = name_quadrature_position()
    return quadrature_preamble('{} = {}.{}({}, {});'.format(name,
                                                            name_paramclass(),
                                                            name,
                                                            intersection,
                                                            pos,
                                                            ),
                               assignees=frozenset({name}),
                               read_variables=frozenset({pos}),
                               )


def define_parameter_temporary(t):
    @preamble
    def _define(name, shape, shape_impl):
        assert len(shape) == 0
        return "{} {};".format(t, name)
    return _define


def cell_parameter_function(name, expr, restriction, t='double'):
    temporary_variable(name, shape=(), decl_method=define_parameter_temporary(t))
    define_parameter_function_class_member(name, expr, t, True)
    evaluate_cell_parameter_function(name, restriction)


def intersection_parameter_function(name, expr, t='double'):
    temporary_variable(name, shape=(), decl_method=define_parameter_temporary(t))
    define_parameter_function_class_member(name, expr, t, False)
    evaluate_intersection_parameter_function(name)
